// SPDX-FileCopyrightText: 2024 Bernardo Gomes Negri <b.gomes.negri@gmail.com>
//
// SPDX-License-Identifier: GPL-2.0-or-later
module;
#include "onedrivedebug.h"
#include <KAccounts/AccountsModel>
#include <KAccounts/ServicesModel>
#include <KAccounts/GetCredentialsJob>
#include <Accounts/Account>
#include <QUrl>
#include <QByteArray>
#include <QDateTime>
#include <QHash>
#include <qloggingcategory.h>
#include "fixedqstringliteral.h"
#include <qvariant.h>
#include <QLoggingCategory>
#include <variant>
#include <tuple>
export module URLHandler;
import URLUtils;

export {
    struct Nothing{};
    
    enum URLError {
        NoSuchAccount,
        NoSuchDrive,
        CredentialsError,
        NotEnoughInfo,
    };

    struct RequestInfo {
        QString bearerToken;
        QUrl url;
    };
}

struct ExpiringToken {
    QString token;
    QDateTime expirationDate;
};

struct AccountDriveName {
    QString accountName;
    QString driveName;
};

size_t qHash(const AccountDriveName &accdrvName, size_t seed = 0) {
    size_t acc = qHash(accdrvName.accountName, seed);
    acc ^= qHash(accdrvName.driveName, acc);
    return acc;
}

bool operator==(const AccountDriveName &a, const AccountDriveName &b) {
    return a.accountName == b.accountName && a.driveName == b.driveName;
}

export class URLHandler {
public:
    URLHandler() :
    m_accModel()
    {
    }
    
    // Returns the URL to the Drive object ("https://graph.microsoft.com/v1.0/me/drive")
    // Do not confuse with the URL to the root folder of the drive
    std::variant<RequestInfo, URLError> driveRequest(const QUrl &specialUri) {
        SpecialUri parsedUri{specialUri};
        if (!parsedUri.isValid) {
            return URLError::NotEnoughInfo;
        }
        auto maybeToken = getTokenForAccName(parsedUri.accountName);
        if (std::holds_alternative<URLError>(maybeToken)) {
            return std::get<URLError>(maybeToken);
        } else {
            QString token = std::get<QString>(maybeToken);
            QUrl driveUrl;
            if (parsedUri.application == Application::OneDrive) {
                driveUrl = QUrl{QStringLiteral("https://graph.microsoft.com/v1.0/me/drive")};
            } else {
                auto maybeDriveId = driveNameToId(parsedUri.accountName, parsedUri.driveName);
                if (!maybeDriveId) {
                    return URLError::NoSuchDrive;
                }
                driveUrl = QUrl{QStringLiteral("https://graph.microsoft.com/v1.0/drives/") + maybeDriveId.value()};
            }
            return RequestInfo{token, driveUrl};
        }
    }

    // Returns the /drives endpoint
    std::variant<RequestInfo, URLError> listDrivesFromWeb(QString accName) {
        return endpointRequest(accName, QStringLiteral("/me/drives"), QString());
    }

    // Gives a token for a certain graph endpoint with a certain account
    std::variant<RequestInfo, URLError> endpointRequest(QString accName, QString graphPath, QString graphQuery) {
        auto maybeToken = getTokenForAccName(accName);
        if (std::holds_alternative<URLError>(maybeToken)) {
            return std::get<URLError>(maybeToken);
        }
        QString token = std::get<QString>(maybeToken);
        QUrl url{QStringLiteral("https://graph.microsoft.com/")};
        url.setPath(QStringLiteral("/v1.0") + graphPath, QUrl::StrictMode);
        url.setQuery(graphQuery, QUrl::StrictMode);
        return RequestInfo {token, url};
    }

    void addDriveToTable(QString accName, QString driveName, QString driveId) {
        m_driveNameIdTable[AccountDriveName{accName, driveName}] = driveId;
    }

    // Converts a onedrive:/ or sharepoint:/ KIO URI to a Onedrive driveItem API address + bearer token
    // See https://learn.microsoft.com/en-us/onedrive/developer/rest-api/concepts/addressing-driveitems
    std::variant<RequestInfo, URLError> specialUriToRequest(const QUrl &specialUri) {
        return specialUriToRequest(SpecialUri{specialUri});
    }

    std::variant<RequestInfo, URLError> specialUriToRequest(const SpecialUri &parsedUri) {
        if (!parsedUri.isValid) {
            return URLError::NotEnoughInfo;
        }
        auto maybeApiUrl = parsedUri.toApiUrl([&](QString accName, QString drvName){return driveNameToId(accName,drvName);});
        if (!maybeApiUrl) {
            return URLError::NoSuchDrive;
        }
        QUrl apiUrl = maybeApiUrl.value();
        QString accName = parsedUri.accountName;
        auto maybeToken = getTokenForAccName(accName);
        if (std::holds_alternative<URLError>(maybeToken)) {
            std::variant<RequestInfo, URLError> ret(std::get<URLError>(maybeToken));
            return ret;
        } else {
            QString token = std::get<QString>(maybeToken);
            RequestInfo reqInfo;
            reqInfo.bearerToken = token;
            reqInfo.url = apiUrl;
            return std::variant<RequestInfo, URLError>(reqInfo);
        }
    }
    
    QList<QString> getAccNames() {
        QList<QString> ret;
        for (int i = 0; i < m_accModel.rowCount(); i++) {
            QModelIndex idx = m_accModel.index(i);
            if (!isAccGood(idx, Nothing())) {
                continue;
            }
            ret.append(m_accModel.data(idx, KAccounts::AccountsModel::DisplayNameRole).toString());
        }
        return ret;
    }

    QList<QString> getDriveNamesForAccount(QString acc) {
        QList<QString> ret;
        for (auto it = m_driveNameIdTable.keyBegin(), end = m_driveNameIdTable.keyEnd(); it != end; ++it) {
            const AccountDriveName &driveAndAcc = *it;
            if (driveAndAcc.accountName == acc) {
                ret.append(driveAndAcc.driveName);
            }
        }
        return ret;
    }

private:
    std::optional<QString> driveNameToId(QString accountName, QString driveName) {
        AccountDriveName key{accountName,driveName};
        return m_driveNameIdTable.contains(key) ? std::make_optional(m_driveNameIdTable[key]) : std::nullopt;
    }

    bool isAccGood(QModelIndex idx, std::variant<QString, Nothing> maybeName) {
        QString provider = m_accModel.data(idx, KAccounts::AccountsModel::ProviderNameRole).toString();
        if (provider != QStringLiteral("microsoft")) {
            return false;
        }
        if (!m_accModel.data(idx, KAccounts::AccountsModel::EnabledRole).toBool()) {
            return false;
        }
        if (std::holds_alternative<QString>(maybeName)) {
            QString accName = std::get<QString>(maybeName);
            QString thisAccName = m_accModel.data(idx, KAccounts::AccountsModel::DisplayNameRole).toString();
            if (thisAccName != accName) {
                return false;
            }
        }
        
        KAccounts::ServicesModel *services = m_accModel.data(idx, KAccounts::AccountsModel::ServicesRole).value<KAccounts::ServicesModel*>();
        for (int i = 0; i < services->rowCount(); i++) {
            QModelIndex servicesModelIndex = services->index(i);
            if (!services->data(servicesModelIndex, KAccounts::ServicesModel::Roles::EnabledRole).toBool()) {
                continue;
            }
            QString serviceName = services->data(servicesModelIndex, KAccounts::ServicesModel::NameRole).toString();
            if (serviceName == QStringLiteral("microsoft-onedrive")) {
                return true;
            }
        }
        return false;
    }
    
    std::variant<QString, URLError> getTokenForAccName(QString accName) {
        if (m_accountTokenCache.contains(accName)) {
            ExpiringToken cachedToken = m_accountTokenCache.value(accName);
            QDateTime now = QDateTime::currentDateTimeUtc();
            if (now < cachedToken.expirationDate) {
                return cachedToken.token;
            }
        }
        
        URLError err = URLError::NoSuchAccount;
        bool has_err = true;
        QString token = QString();
        for (int i = 0; i < m_accModel.rowCount(); i++) {
            QModelIndex modelIndex = m_accModel.index(i);
            if (!isAccGood(modelIndex, accName)) {
                continue;
            }
            // We have found our account! Now get that bearer token
            has_err = false;
            auto job = KAccounts::GetCredentialsJob(m_accModel.data(modelIndex, KAccounts::AccountsModel::IdRole).value<Accounts::AccountId>(), nullptr);
            if (!job.exec()) {
                qCWarning(ONEDRIVE) << "Error with GetCredentialsJob: " << job.errorString();
                err = URLError::CredentialsError;
                has_err = true;
            } else {
                QVariantMap data = job.credentialsData();
                token = data.value(QStringLiteral("AccessToken")).toString();
                ExpiringToken expToken;
                expToken.token = token;
                // We give ourselves a few seconds of leeway
                expToken.expirationDate = QDateTime::currentDateTimeUtc().addSecs(data.value(QStringLiteral("ExpiresIn")).toInt() - 20);
                qCDebug(ONEDRIVE) << "Token for account" << accName << "expires at" << expToken.expirationDate;
                m_accountTokenCache[accName] = expToken;
                // The access token is a secret (one that expires in ~15 minutes)
                // Only uncomment this line if you wish to experiment with the Onedrive API.
                // Debug lines go to journalctl!!
                //qCDebug(ONEDRIVE) << "token=" << token;
            }
            break;
        }
        if (has_err) {
            return std::variant<QString, URLError>(err);
        } else {
            return std::variant<QString, URLError>(token);
        }
    }

    KAccounts::AccountsModel m_accModel;
    QHash<QString, ExpiringToken> m_accountTokenCache;
    QHash<AccountDriveName,QString> m_driveNameIdTable;
};
