// SPDX-FileCopyrightText: 2024 Bernardo Gomes Negri <b.gomes.negri@gmail.com>
//
// SPDX-License-Identifier: GPL-2.0-or-later

module;
#include <QUrl>
#include <QString>
#include "fixedqstringliteral.h"
#include <tuple>
export module URLUtils;

export enum Application {
    OneDrive,
    Sharepoint
};

export Application schemeToApplication(QString scheme) {
    return scheme == QStringLiteral("onedrive") ? Application::OneDrive : Application::Sharepoint;
}

// Leaves a trailing slash
export QUrl resolveUrl(const QUrl &url) {
    // Resolve onedrive:/account/folder1/../folder2 to onedrive:/account/folder2/
    QUrl ret = url;
    QString inputPath = ret.path(QUrl::FullyDecoded);
    if (!inputPath.endsWith(QStringLiteral("/"))) {
        ret.setPath(inputPath + QStringLiteral("/"));
    }
    return ret.resolved(QUrl(QStringLiteral(".")));
}

export class SpecialUri {
public:
    // isValid only refers to if the URI could correspond to an actual folder or file
    // on the cloud.
    bool isValid;
    Application application;
    QString accountName;
    QString driveName;
    QString path;

    explicit SpecialUri(const QUrl &specialUri) :
    isValid(false),
    application(schemeToApplication(specialUri.scheme()))
    {
        // Expects either onedrive:/account/path or sharepoint:/account/drive/path
        QUrl resolvedUri = resolveUrl(specialUri);
        QStringList splitPath = resolvedUri.path(QUrl::FullyDecoded).split(QStringLiteral("/"));
        int pathSize = splitPath.size();
        if (pathSize < 3) {
            // No account name info
            return;
        }
        accountName = splitPath.at(1);
        if (application == Application::Sharepoint) {
            if (pathSize < 4) {
                // No drive name info
                return;
            }
            driveName = splitPath.at(2);
            path = splitPath.mid(3).join(QStringLiteral("/"));
        } else {
            path = splitPath.mid(2).join(QStringLiteral("/"));
        }
        if (path.size() == 0 || path.at(0) != u'/') {
            path = path.prepend(QStringLiteral("/"));
        }
        isValid = true;
    }

    std::optional<QUrl> toApiUrl(std::optional<std::function<std::optional<QString>(QString,QString)>> driveNameToId = std::nullopt) const {
        // See https://learn.microsoft.com/en-us/onedrive/developer/rest-api/concepts/addressing-driveitems
        if (!isValid) {
            return std::nullopt;
        }
        QUrl ret;
        ret.setScheme(QStringLiteral("https"));
        ret.setHost(QStringLiteral("graph.microsoft.com"));
        QString drivePath = path;
        if (drivePath.endsWith(QStringLiteral("/"))) {
            drivePath = drivePath.chopped(1);
        }
        if (application == Application::OneDrive) {
            if (drivePath.isEmpty()) {
                drivePath = QStringLiteral("/me/drive/root");
            } else {
                drivePath = QStringLiteral("/me/drive/root:") + drivePath + QStringLiteral(":");
            }
        } else {
            if (!driveNameToId) {
                return std::nullopt;
            }
            std::optional<QString> maybeDriveId = driveNameToId.value()(accountName, driveName);
            if (!maybeDriveId) {
                return std::nullopt;
            }
            QString driveId = maybeDriveId.value();
            if (drivePath.isEmpty()) {
                drivePath = QStringLiteral("/drives/") + driveId + QStringLiteral("/root");
            } else {
                drivePath = QStringLiteral("/drives/") + driveId + QStringLiteral("/root:") + drivePath + QStringLiteral(":");
            }
        }
        ret.setPath(QStringLiteral("/v1.0") + drivePath, QUrl::DecodedMode);
        return ret;
    }

    // Is this something that has an actual name on the cloud?
    bool isAddressable() const {
        return isValid && path != QStringLiteral("/");
    }
};

export bool isAddressable(const QUrl &url) {
    return SpecialUri{url}.isAddressable();
}

// Can we move stuff into this?
export bool atLeastRoot(const QUrl &url) {
    return SpecialUri{url}.isValid;
}

export int urlSize(const QUrl &url) {
    return resolveUrl(url).path(QUrl::FullyDecoded).split(QStringLiteral("/")).size();
}

// I think the parts of the URL path are actually called fragments
export std::tuple<QString,QString> splitLastPart(QString path) {
    QStringList splitPath = path.split(QStringLiteral("/"));
    if (splitPath.size() < 3) {
        return {path,QString()};
    }
    splitPath.removeLast();
    QString lastPart = splitPath.takeLast();
    QString newPath = splitPath.join(QStringLiteral("/"));
    if (newPath.isEmpty() || newPath.at(0) != u'/') {
        newPath = QStringLiteral("/");
    }
    return {newPath,lastPart};
}


export std::tuple<QUrl, QString> splitLastPart(const QUrl &url) {
    QUrl newUrl = resolveUrl(url);
    auto splitPath = splitLastPart(newUrl.path(QUrl::FullyDecoded));
    newUrl.setPath(std::get<0>(splitPath), QUrl::DecodedMode);
    return std::tuple<QUrl, QString>(newUrl, std::get<1>(splitPath));
}

export bool testEnv(const char *envVarName) {
    QString var = qEnvironmentVariable(envVarName, QStringLiteral("0"));
    return var == QStringLiteral("1") || var == QStringLiteral("true");
}
